@@ -1,13 +1,15 @@
+/*
+  Copyright (c) 1990-1999 Info-ZIP.  All rights reserved.
+
+  See the accompanying file LICENSE, version 1999-Oct-05 or later
+  (the contents of which are also included in zip.h) for terms of use.
+  If, for some reason, both of these files are missing, the Info-ZIP license
+  also may be found at:  ftp://ftp.cdrom.com/pub/infozip/license.html
+*/
 /*
 
- Copyright (C) 1990-1997 Mark Adler, Richard B. Wales, Jean-loup Gailly,
- Kai Uwe Rommel, Onno van der Linden, Igor Mandrichenko and Chris Herborth.
- Permission is granted to any individual or institution to use, copy, or
- redistribute this software so long as all of the original files are included,
- that it is not sold for profit, and that this copyright notice is retained.
-
  This BeOS-specific file is based on unix.c in the unix directory; changes
- by Chris Herborth (chrish@qnx.com).
+ by Chris Herborth (chrish@pobox.com).
 
 */
 
@@ -25,20 +27,12 @@
 #include <unistd.h>
 #include <string.h>
 
-#define MATCH shmatch
-
 #include <dirent.h>
 
-#ifdef __GNUC__
-#warning NOTE: Expect two warnings for Mime.h...
-#endif
 #include <kernel/fs_attr.h>
 #include <storage/Mime.h>
+#include <support/byteorder.h>
 
-int get_attr_dir( const char *, char **, off_t * );
-int add_UT_ef( struct zlist far * );
-int add_Ux_ef( struct zlist far * );
-int add_Be_ef( struct zlist far * );
 
 #define PAD 0
 #define PATH_END '/'
@@ -58,6 +52,10 @@ local time_t label_utim = 0;
 
 /* Local functions */
 local char *readd OF((DIR *));
+local int get_attr_dir( const char *, char **, off_t * );
+local int add_UT_ef( struct zlist far * );
+local int add_Ux_ef( struct zlist far * );
+local int add_Be_ef( struct zlist far * );
 
 
 #ifdef NO_DIR                    /* for AT&T 3B1 */
@@ -102,8 +100,9 @@ DIR *d;                 /* directory stream to read from */
   return e == NULL ? (char *) NULL : e->d_name;
 }
 
-int procname(n)
+int procname(n, caseflag)
 char *n;                /* name to process */
+int caseflag;           /* true to force case-sensitive match */
 /* Process a name or sh expression to operate on (or exclude).  Return
    an error code in the ZE_ class. */
 {
@@ -116,16 +115,16 @@ char *n;                /* name to process */
   struct zlist far *z;  /* steps through zfiles list */
 
   if (strcmp(n, "-") == 0)   /* if compressing stdin */
-    return newname(n, 0);
+    return newname(n, 0, caseflag);
   else if (LSSTAT(n, &s))
   {
     /* Not a file or directory--search for shell expression in zip file */
     p = ex2in(n, 0, (int *)NULL);       /* shouldn't affect matching chars */
     m = 1;
     for (z = zfiles; z != NULL; z = z->nxt) {
-      if (MATCH(p, z->iname))
+      if (MATCH(p, z->iname, caseflag))
       {
-        z->mark = pcount ? filter(z->zname) : 1;
+        z->mark = pcount ? filter(z->zname, caseflag) : 1;
         if (verbose)
             fprintf(mesg, "zip diagnostic: %scluding %s\n",
                z->mark ? "in" : "ex", z->name);
@@ -140,7 +139,7 @@ char *n;                /* name to process */
   if ((s.st_mode & S_IFDIR) == 0)
   {
     /* add or remove name of file */
-    if ((m = newname(n, 0)) != ZE_OK)
+    if ((m = newname(n, 0, caseflag)) != ZE_OK)
       return m;
   } else {
     /* Add trailing / to the directory name */
@@ -153,7 +152,7 @@ char *n;                /* name to process */
       a = p + strlen(p);
       if (a[-1] != '/')
         strcpy(a, "/");
-      if (dirnames && (m = newname(p, 1)) != ZE_OK) {
+      if (dirnames && (m = newname(p, 1, caseflag)) != ZE_OK) {
         free((zvoid *)p);
         return m;
       }
@@ -171,7 +170,7 @@ char *n;                /* name to process */
             return ZE_MEM;
           }
           strcat(strcpy(a, p), e);
-          if ((m = procname(a)) != ZE_OK)   /* recurse on name */
+          if ((m = procname(a, caseflag)) != ZE_OK)   /* recurse on name */
           {
             if (m == ZE_MISS)
               zipwarn("name not matched: ", a);
@@ -203,7 +202,9 @@ int *pdosflag;          /* output: force MSDOS file attributes? */
 
   /* Find starting point in name before doing malloc */
   for (t = x; *t == '/'; t++)
-    ;
+    ;                   /* strip leading '/' chars to get a relative path */
+  while (*t == '.' && t[1] == '/')
+    t += 2;             /* strip redundant leading "./" sections */
 
   /* Make changes, if any, to the copied name (leave original intact) */
   if (!pathput)
@@ -214,7 +215,7 @@ int *pdosflag;          /* output: force MSDOS file attributes? */
     return NULL;
   strcpy(n, t);
 
-  if (isdir == 42) return n;      /* avoid warning on unused variable */
+  if (isdir == 42) return n;    /* avoid warning on unused variable */
 
   if (dosify)
     msname(n);
@@ -240,7 +241,7 @@ char *n;                /* internal file name */
 }
 
 /*
- * use ztimbuf in both POSIX and non POSIX cases ?
+ * XXX use ztimbuf in both POSIX and non POSIX cases ?
  */
 void stamp(f, d)
 char *f;                /* name of file to change */
@@ -355,13 +356,11 @@ int get_attr_dir( const char *name, char **attr_buff, off_t *total_size )
     int               fd;
     DIR              *fa_dir;
     struct dirent    *fa_ent;
-#ifdef __GNUC__
-#warning Expect a warning for attrs_size... gcc is lying.
-#endif
     off_t             attrs_size;
     off_t             this_size;
     char             *ptr;
     struct attr_info  fa_info;
+    struct attr_info  big_fa_info;
 
     retval      = EOK;
     attrs_size  = 0;    /* gcc still says this is used uninitialized... */
@@ -375,7 +374,16 @@ int get_attr_dir( const char *name, char **attr_buff, off_t *total_size )
 
     /* ----------------------------------------------------------------- */
     /* Can we open the file/directory?                                   */
-    fd = open( name, O_RDONLY );
+    /*                                                                   */
+    /* linkput is a zip global; it's set to 1 if we're storing symbolic  */
+    /* links as symbolic links (instead of storing the thing the link    */
+    /* points to)... if we're storing the symbolic link as a link, we'll */
+    /* want the link's file attributes, otherwise we want the target's.  */
+    if( linkput ) {
+        fd = open( name, O_RDONLY | O_NOTRAVERSE );
+    } else {
+        fd = open( name, O_RDONLY );
+    }
     if( fd < 0 ) {
         return errno;
     }
@@ -436,7 +444,11 @@ int get_attr_dir( const char *name, char **attr_buff, off_t *total_size )
         ptr += strlen( fa_ent->d_name );
         *ptr++ = '\0';
 
-        memcpy( ptr, &fa_info, sizeof( struct attr_info ) );
+        /* We need to put a big-endian version of the fa_info data into */
+        /* the archive.                                                 */
+        big_fa_info.type = B_HOST_TO_BENDIAN_INT32( fa_info.type );
+        big_fa_info.size = B_HOST_TO_BENDIAN_INT64( fa_info.size );
+        memcpy( ptr, &big_fa_info, sizeof( struct attr_info ) );
         ptr += sizeof( struct attr_info );
 
         if( fa_info.size > 0 ) {
@@ -452,6 +464,12 @@ int get_attr_dir( const char *name, char **attr_buff, off_t *total_size )
                          (ssize_t)read_bytes, (ssize_t)fa_info.size );
                 zipwarn( "attribute size mismatch: ", buff );
             }
+
+            /* Wave my magic wand... this swaps all the Be types to big- */
+            /* endian automagically.                                     */
+            (void)swap_data( fa_info.type, ptr, fa_info.size,
+                             B_SWAP_HOST_TO_BENDIAN );
+
             ptr += fa_info.size;
         }
 
@@ -488,12 +506,17 @@ int get_attr_dir( const char *name, char **attr_buff, off_t *total_size )
 #define EB_L_UT_SIZE    (EB_HEADSIZE + EB_UT_LEN(2))
 #define EB_C_UT_SIZE    (EB_HEADSIZE + EB_UT_LEN(1))
 
-int add_UT_ef( struct zlist far *z )
+local int add_UT_ef( struct zlist far *z )
 {
     char        *l_ef = NULL;
     char        *c_ef = NULL;
     struct stat  s;
 
+#ifdef IZ_CHECK_TZ
+    if (!zp_tz_is_valid)
+        return ZE_OK;           /* skip silently if no valid TZ info */
+#endif
+
     /* We can't work if there's no entry to work on. */
     if( z == NULL ) {
         return ZE_LOGIC;
@@ -513,34 +536,34 @@ int add_UT_ef( struct zlist far *z )
 
     /* Allocate memory for the local and central extra fields. */
     if( z->extra && z->ext != 0 ) {
-        z->extra = (char *)realloc( z->extra, z->ext + EB_L_UT_SIZE );
-        l_ef = z->extra + z->ext;
+        l_ef = (char *)realloc( z->extra, z->ext + EB_L_UT_SIZE );
     } else {
-        z->extra = (char *)malloc( EB_L_UT_SIZE );
+        l_ef = (char *)malloc( EB_L_UT_SIZE );
         z->ext = 0;
-        l_ef = z->extra;
     }
     if( l_ef == NULL ) {
         return ZE_MEM;
     }
+    z->extra = l_ef;
+    l_ef += z->ext;
 
     if( z->cextra && z->cext != 0 ) {
-        z->cextra = (char *)realloc( z->cextra, z->cext + EB_C_UT_SIZE );
-        c_ef = z->cextra + z->cext;
+        c_ef = (char *)realloc( z->cextra, z->cext + EB_C_UT_SIZE );
     } else {
-        z->cextra = (char *)malloc( EB_C_UT_SIZE );
+        c_ef = (char *)malloc( EB_C_UT_SIZE );
         z->cext = 0;
-        c_ef = z->cextra;
     }
     if( c_ef == NULL ) {
         return ZE_MEM;
     }
+    z->cextra = c_ef;
+    c_ef += z->cext;
 
     /* Now add the local version of the field. */
     *l_ef++ = 'U';
     *l_ef++ = 'T';
     *l_ef++ = (char)(EB_UT_LEN(2)); /* length of data in local EF */
-    *l_ef++ = (char)(EB_UT_LEN(2) >> 8);
+    *l_ef++ = (char)0;
     *l_ef++ = (char)(EB_UT_FL_MTIME | EB_UT_FL_ATIME);
     *l_ef++ = (char)(s.st_mtime);
     *l_ef++ = (char)(s.st_mtime >> 8);
@@ -554,15 +577,8 @@ int add_UT_ef( struct zlist far *z )
     z->ext += EB_L_UT_SIZE;
 
     /* Now add the central version. */
-    *c_ef++ = 'U';
-    *c_ef++ = 'T';
-    *c_ef++ = (char)(EB_UT_LEN(1)); /* length of data in central EF */
-    *c_ef++ = (char)(EB_UT_LEN(1) >> 8);
-    *c_ef++ = (char)(EB_UT_FL_MTIME);
-    *c_ef++ = (char)(s.st_mtime);
-    *c_ef++ = (char)(s.st_mtime >> 8);
-    *c_ef++ = (char)(s.st_mtime >> 16);
-    *c_ef++ = (char)(s.st_mtime >> 24);
+    memcpy(c_ef, l_ef-EB_L_UT_SIZE, EB_C_UT_SIZE);
+    c_ef[EB_LEN] = (char)(EB_UT_LEN(1)); /* length of data in central EF */
 
     z->cext += EB_C_UT_SIZE;
 
@@ -575,7 +591,7 @@ int add_UT_ef( struct zlist far *z )
 #define EB_L_UX2_SIZE   (EB_HEADSIZE + EB_UX2_MINLEN)
 #define EB_C_UX2_SIZE   (EB_HEADSIZE)
 
-int add_Ux_ef( struct zlist far *z )
+local int add_Ux_ef( struct zlist far *z )
 {
     char        *l_ef = NULL;
     char        *c_ef = NULL;
@@ -595,28 +611,28 @@ int add_Ux_ef( struct zlist far *z )
 
     /* Allocate memory for the local and central extra fields. */
     if( z->extra && z->ext != 0 ) {
-        z->extra = (char *)realloc( z->extra, z->ext + EB_L_UX2_SIZE );
-        l_ef = z->extra + z->ext;
+        l_ef = (char *)realloc( z->extra, z->ext + EB_L_UX2_SIZE );
     } else {
-        z->extra = (char *)malloc( EB_L_UX2_SIZE );
+        l_ef = (char *)malloc( EB_L_UX2_SIZE );
         z->ext = 0;
-        l_ef = z->extra;
     }
     if( l_ef == NULL ) {
         return ZE_MEM;
     }
+    z->extra = l_ef;
+    l_ef += z->ext;
 
     if( z->cextra && z->cext != 0 ) {
-        z->cextra = (char *)realloc( z->cextra, z->cext + EB_C_UX2_SIZE );
-        c_ef = z->cextra + z->cext;
+        c_ef = (char *)realloc( z->cextra, z->cext + EB_C_UX2_SIZE );
     } else {
-        z->cextra = (char *)malloc( EB_C_UX2_SIZE );
+        c_ef = (char *)malloc( EB_C_UX2_SIZE );
         z->cext = 0;
-        c_ef = z->cextra;
     }
     if( c_ef == NULL ) {
         return ZE_MEM;
     }
+    z->cextra = c_ef;
+    c_ef += z->cext;
 
     /* Now add the local version of the field. */
     *l_ef++ = 'U';
@@ -647,9 +663,11 @@ int add_Ux_ef( struct zlist far *z )
 #define EB_L_BE_SIZE    (EB_HEADSIZE + EB_L_BE_LEN) /* + attr size */
 #define EB_C_BE_SIZE    (EB_HEADSIZE + EB_C_BE_LEN)
 
-#define MEMCOMPRESS_OVERHEAD    6   /* ush compression type, ulg CRC */
+#define MEMCOMPRESS_HEADER      6   /* ush compression type, ulg CRC */
+#define DEFLAT_WORSTCASE_ADD    5   /* byte blocktype, 2 * ush blocklength */
+#define MEMCOMPRESS_OVERHEAD    (MEMCOMPRESS_HEADER + DEFLAT_WORSTCASE_ADD)
 
-int add_Be_ef( struct zlist far *z )
+local int add_Be_ef( struct zlist far *z )
 {
     char *l_ef       = NULL;
     char *c_ef       = NULL;
@@ -659,22 +677,6 @@ int add_Be_ef( struct zlist far *z )
     ush   compsize   = 0;
     uch   flags      = 0;
 
-#ifdef BE_NO_SYMLINK_ATTRS
-    /* See osdep.h; there's currently no way to save/restore a symbolic     */
-    /* link's attributes from C, and I'm loathe to introduce any C++. [cjh] */
-    {
-        struct stat s;
-
-        if( SSTAT( z->name, &s ) ) {
-            return ZE_OPEN;
-        }
-
-        if( S_ISLNK( s.st_mode ) ) {
-            return ZE_OK;
-        }
-    }
-#endif
-
     /* Check to make sure we've got enough room in the extra fields. */
     if( z->ext + EB_L_BE_SIZE > USHRT_MAX ||
         z->cext + EB_C_BE_SIZE > USHRT_MAX ) {
@@ -704,7 +706,7 @@ int add_Be_ef( struct zlist far *z )
     }
 
     if( verbose ) {
-        printf( "\t[in=%ld]", (unsigned long)attrsize );
+        printf( "\t[in=%lu]", (unsigned long)attrsize );
     }
 
     /* Try compressing the data */
@@ -717,7 +719,7 @@ int add_Be_ef( struct zlist far *z )
                             attrbuff,
                             (size_t)attrsize );
     if( verbose ) {
-        printf( " [out=%d]", compsize );
+        printf( " [out=%u]", compsize );
     }
 
     /* Attempt to optimise very small attributes. */
@@ -736,28 +738,28 @@ int add_Be_ef( struct zlist far *z )
 
     /* Allocate memory for the local and central extra fields. */
     if( z->extra && z->ext != 0 ) {
-        z->extra = (char *)realloc( z->extra, z->ext + EB_L_BE_SIZE + compsize );
-        l_ef = z->extra + z->ext;
+        l_ef = (char *)realloc( z->extra, z->ext + EB_L_BE_SIZE + compsize );
     } else {
-        z->extra = (char *)malloc( EB_L_BE_SIZE + compsize );
+        l_ef = (char *)malloc( EB_L_BE_SIZE + compsize );
         z->ext = 0;
-        l_ef = z->extra;
     }
     if( l_ef == NULL ) {
         return ZE_MEM;
     }
+    z->extra = l_ef;
+    l_ef += z->ext;
 
     if( z->cextra && z->cext != 0 ) {
-        z->cextra = (char *)realloc( z->cextra, z->cext + EB_C_BE_SIZE );
-        c_ef = z->cextra + z->cext;
+        c_ef = (char *)realloc( z->cextra, z->cext + EB_C_BE_SIZE );
     } else {
-        z->cextra = (char *)malloc( EB_C_BE_SIZE );
+        c_ef = (char *)malloc( EB_C_BE_SIZE );
         z->cext = 0;
-        c_ef = z->cextra;
     }
     if( c_ef == NULL ) {
         return ZE_MEM;
     }
+    z->cextra = c_ef;
+    c_ef += z->cext;
 
     /* Now add the local version of the field. */
     *l_ef++ = 'B';
@@ -914,6 +916,12 @@ void version_local()
 
 #ifdef __POWERPC__
       " (PowerPC)",
+#else
+#  ifdef __INTEL__
+      " (x86)",
+#  else
+      " (UNKNOWN!)",
+#  endif
 #endif
 
 #ifdef __DATE__
