@@ -1,18 +1,16 @@
 /*
+  Copyright (c) 1990-1999 Info-ZIP.  All rights reserved.
 
- Copyright (C) 1990-1997 Mark Adler, Richard B. Wales, Jean-loup Gailly,
- Kai Uwe Rommel, Onno van der Linden and Igor Mandrichenko.
- Permission is granted to any individual or institution to use, copy, or
- redistribute this software so long as all of the original files are included,
- that it is not sold for profit, and that this copyright notice is retained.
-
+  See the accompanying file LICENSE, version 1999-Oct-05 or later
+  (the contents of which are also included in zip.h) for terms of use.
+  If, for some reason, both of these files are missing, the Info-ZIP license
+  also may be found at:  ftp://ftp.cdrom.com/pub/infozip/license.html
 */
-
 /*
  *  zipfile.c by Mark Adler.
  */
+#define __ZIPFILE_C
 
-#define NOCPYRT         /* this is not a main module */
 #include "zip.h"
 #include "revision.h"
 
@@ -23,12 +21,22 @@
 #  include "vms/vmsdefs.h"
 #endif
 
+#ifdef __RSXNT__
+#  include <windows.h>
+#endif
+
 /*
  * XXX start of zipfile.h
  */
+#ifdef THEOS
+/* Macros cause stack overflow in compiler */
+ush SH(uch* p) { return ((ush)(uch)((p)[0]) | ((ush)(uch)((p)[1]) << 8)); }
+ulg LG(uch* p) { return ((ulg)(SH(p)) | ((ulg)(SH((p)+2)) << 16)); }
+#else /* !THEOS */
 /* Macros for converting integers in little-endian to machine format */
-#define SH(a) (((ush)(uch)(a)[0]) | (((ush)(uch)(a)[1]) << 8))
+#define SH(a) ((ush)(((ush)(uch)(a)[0]) | (((ush)(uch)(a)[1]) << 8)))
 #define LG(a) ((ulg)SH(a) | ((ulg)SH((a)+2) << 16))
+#endif /* ?THEOS */
 
 /* Macros for writing machine integers to little-endian format */
 #define PUTSH(a,f) {putc((char)((a) & 0xff),(f)); putc((char)((a) >> 8),(f));}
@@ -89,16 +97,17 @@
 /* Local functions */
 
 local int zqcmp OF((ZCONST zvoid *, ZCONST zvoid *));
+local int scanzipf_reg OF((FILE *f));
 #ifndef UTIL
-   local void zipoddities OF((struct zlist far *));
-   local int readzipfile2 OF((void));
    local int rqcmp OF((ZCONST zvoid *, ZCONST zvoid *));
    local int zbcmp OF((ZCONST zvoid *, ZCONST zvoid far *));
+   local void zipoddities OF((struct zlist far *));
+   local int scanzipf_fix OF((FILE *f));
 #  ifdef USE_EF_UT_TIME
      local int ef_scan_ut_time OF((char *ef_buf, extent ef_len, int ef_is_cent,
                                    iztimes *z_utim));
 #  endif /* USE_EF_UT_TIME */
-   local void cutpath OF((char *p));
+   local void cutpath OF((char *p, int delim));
 #endif /* !UTIL */
 
 /*
@@ -143,13 +152,14 @@ ZCONST zvoid far *z;    /* pointer to a pointer to a zip entry */
 
 
 struct zlist far *zsearch(n)
-char *n;                /* name to find */
+ZCONST char *n;         /* name to find */
 /* Return a pointer to the entry in zfile with the name n, or NULL if
    not found. */
 {
   zvoid far **p;        /* result of search() */
 
-  if (zcount && (p = search(n, (zvoid far **)zsort, zcount, zbcmp)) != NULL)
+  if (zcount &&
+      (p = search(n, (ZCONST zvoid far **)zsort, zcount, zbcmp)) != NULL)
     return *(struct zlist far **)p;
   else
     return NULL;
@@ -158,7 +168,9 @@ char *n;                /* name to find */
 #endif /* !UTIL */
 
 #ifndef VMS
-#  define PATHCUT '/'
+#  ifndef PATHCUT
+#    define PATHCUT '/'
+#  endif
 
 char *ziptyp(s)
 char *s;                /* file name to force to zip */
@@ -170,6 +182,10 @@ char *s;                /* file name to force to zip */
 {
   char *q;              /* temporary pointer */
   char *t;              /* pointer to malloc'ed string */
+#ifdef THEOS
+  char *r;              /* temporary pointer */
+  char *disk;
+#endif
 
   if ((t = malloc(strlen(s) + 5)) == NULL)
     return NULL;
@@ -178,35 +194,65 @@ char *s;                /* file name to force to zip */
   _toslash(t);
 #endif
 #ifdef MSDOS
-  for (q = t; *q; q++)
+  for (q = t; *q; INCSTR(q))
     if (*q == '\\')
       *q = '/';
 #endif /* MSDOS */
+#ifdef __RSXNT__   /* RSXNT/EMX C rtl uses OEM charset */
+  AnsiToOem(t, t);
+#endif
   if (adjust) return t;
 #ifndef RISCOS
 # ifndef QDOS
 #  ifdef AMIGA
-  if ((q = strrchr(t, '/')) == NULL)
-    q = strrchr(t, ':');
-  if (strrchr((q ? q + 1 : t), '.') == NULL)
+  if ((q = MBSRCHR(t, '/')) == NULL)
+    q = MBSRCHR(t, ':');
+  if (MBSRCHR((q ? q + 1 : t), '.') == NULL)
 #  else /* !AMIGA */
-#    ifdef TANDEM
-  if (strrchr((q = strrchr(t, '.')) == NULL ? t : q + 1, ' ') == NULL)
-#    else /* !TANDEM */
-  if (strrchr((q = strrchr(t, PATHCUT)) == NULL ? t : q + 1, '.') == NULL)
-#    endif /* TANDEM */
+#    ifdef THEOS
+  /* the argument expansion add a dot to the end of file names when
+   * there is no extension and at least one of a argument has wild cards.
+   * So check for at least one character in the extension if there is a dot
+   * in file name */
+  if ((q = MBSRCHR((q = MBSRCHR(t, PATHCUT)) == NULL ? t : q + 1, '.')) == NULL
+    || q[1] == '\0') {
+#    else /* !THEOS */
+#      ifdef TANDEM
+  if (MBSRCHR((q = MBSRCHR(t, '.')) == NULL ? t : q + 1, ' ') == NULL)
+#      else /* !TANDEM */
+  if (MBSRCHR((q = MBSRCHR(t, PATHCUT)) == NULL ? t : q + 1, '.') == NULL)
+#      endif /* ?TANDEM */
+#    endif /* ?THEOS */
 #  endif /* ?AMIGA */
 #  ifdef CMS_MVS
-     if (strncmp(t,"dd:",3) != 0 && strncmp(t,"DD:",3) != 0)
+    if (strncmp(t,"dd:",3) != 0 && strncmp(t,"DD:",3) != 0)
 #  endif /* CMS_MVS */
-#  ifdef TANDEM       /*  Tandem can't cope with extensions */
+#  ifdef THEOS
+    /* insert .zip extension before disk name */
+    if ((r = MBSRCHR(t, ':')) != NULL) {
+        /* save disk name */
+        if ((disk = strdup(r)) == NULL)
+            return NULL;
+        strcpy(r[-1] == '.' ? r - 1 : r, ".zip");
+        strcat(t, disk);
+        free(disk);
+    } else {
+        if (q != NULL && *q == '.')
+          strcpy(q, ".zip");
+        else
+          strcat(t, ".zip");
+    }
+  }
+#  else /* !THEOS */
+#    ifdef TANDEM     /*  Tandem can't cope with extensions */
     strcat(t, " ZIP");
-#  else /* !TANDEM */
+#    else /* !TANDEM */
     strcat(t, ".zip");
-#  endif /* !TANDEM */
-# else
+#    endif /* ?TANDEM */
+#  endif /* ?THEOS */
+# else /* QDOS */
   q = LastDir(t);
-  if(strrchr(q, '_') == NULL && strrchr(q, '.') == NULL)
+  if(MBSRCHR(q, '_') == NULL && MBSRCHR(q, '.') == NULL)
   {
       strcat(t, "_zip");
   }
@@ -276,130 +322,106 @@ char *s;
 local void zipoddities(z)
 struct zlist far *z;
 {
-   if ((z->vem >> 8) >= NUM_HOSTS)
-   {
-     sprintf(errbuf, "made by version %d.%d on system type %d: ",
-             (ush)(z->vem & 0xff) / (ush)10, (ush)(z->vem & 0xff) % (ush)10,
-             z->vem >> 8);
-     zipwarn(errbuf, z->zname);
-   }
-   if (z->ver != 10 && z->ver != 11 && z->ver != 20)
-   {
-       sprintf(errbuf, "needs unzip %d.%d on system type %d: ",
-               (ush)(z->ver & 0xff) / (ush)10,
-               (ush)(z->ver & 0xff) % (ush)10, z->ver >> 8);
-       zipwarn(errbuf, z->zname);
+    if ((z->vem >> 8) >= NUM_HOSTS)
+    {
+        sprintf(errbuf, "made by version %d.%d on system type %d: ",
+                (ush)(z->vem & 0xff) / (ush)10, (ush)(z->vem & 0xff) % (ush)10,
+                z->vem >> 8);
+        zipwarn(errbuf, z->zname);
+    }
+    if (z->ver != 10 && z->ver != 11 && z->ver != 20)
+    {
+        sprintf(errbuf, "needs unzip %d.%d on system type %d: ",
+                (ush)(z->ver & 0xff) / (ush)10,
+                (ush)(z->ver & 0xff) % (ush)10, z->ver >> 8);
+        zipwarn(errbuf, z->zname);
     }
     if (z->flg != z->lflg)
     {
-       sprintf(errbuf, "local flags = 0x%04x, central = 0x%04x: ",
-               z->lflg, z->flg);
-       zipwarn(errbuf, z->zname);
-     }
-     else if (z->flg & ~0xf)
-     {
-       sprintf(errbuf, "undefined bits used in flags = 0x%04x: ", z->flg);
-       zipwarn(errbuf, z->zname);
-     }
-     if (z->how > DEFLATE)
-     {
-       sprintf(errbuf, "unknown compression method %u: ", z->how);
-       zipwarn(errbuf, z->zname);
-     }
-     if (z->dsk)
-     {
-       sprintf(errbuf, "starts on disk %u: ", z->dsk);
-       zipwarn(errbuf, z->zname);
-      }
-      if (z->att!=ASCII && z->att!=BINARY && z->att!=__EBCDIC)
-      {
+        sprintf(errbuf, "local flags = 0x%04x, central = 0x%04x: ",
+                z->lflg, z->flg);
+        zipwarn(errbuf, z->zname);
+    }
+    else if (z->flg & ~0xf)
+    {
+        sprintf(errbuf, "undefined bits used in flags = 0x%04x: ", z->flg);
+        zipwarn(errbuf, z->zname);
+    }
+    if (z->how > DEFLATE)
+    {
+        sprintf(errbuf, "unknown compression method %u: ", z->how);
+        zipwarn(errbuf, z->zname);
+    }
+    if (z->dsk)
+    {
+        sprintf(errbuf, "starts on disk %u: ", z->dsk);
+        zipwarn(errbuf, z->zname);
+    }
+    if (z->att!=ASCII && z->att!=BINARY && z->att!=__EBCDIC)
+    {
         sprintf(errbuf, "unknown internal attributes = 0x%04x: ", z->att);
         zipwarn(errbuf, z->zname);
-      }
+    }
 #if 0
 /* This test is ridiculous, it produces an error message for almost every */
 /* platform of origin other than MS-DOS, Unix, VMS, and Acorn!  Perhaps   */
 /* we could test "if (z->dosflag && z->atx & ~0xffL)", but what for?      */
-     if (((n = z->vem >> 8) != 3) && n != 2 && n != 13 && z->atx & ~0xffL)
-     {
-       sprintf(errbuf, "unknown external attributes = 0x%08lx: ", z->atx);
-       zipwarn(errbuf, z->zname);
-     }
-#endif
-     if (z->ext || z->cext)
-       if (z->ext && z->cext && z->extra != z->cextra)
-       {
-         sprintf(errbuf,
-                 "local extra (%ld bytes) != central extra (%ld bytes): ",
-                 (ulg)z->ext, (ulg)z->cext);
-          zipwarn(errbuf, z->zname);
-       }
-#ifndef RISCOS
-       else if (noisy)
-#else /* RISCOS */
-/* avoid warnings on zipfiles created on RISCOS itself! */
-/* probably this can be made more generic with z->vem != OS_CODE */
-/* or, was this warning really intended (eg. OS/2)? */
-       else if (noisy && ((z->vem >> 8) != 13))
+    if (((n = z->vem >> 8) != 3) && n != 2 && n != 13 && z->atx & ~0xffL)
+    {
+        sprintf(errbuf, "unknown external attributes = 0x%08lx: ", z->atx);
+        zipwarn(errbuf, z->zname);
+    }
 #endif
+    if (z->ext || z->cext)
+    {
+        if (z->ext && z->cext && z->extra != z->cextra)
         {
-        fprintf(stderr, "zip info: %s has %ld bytes of %sextra data\n",
-                z->zname, z->ext ? (ulg)z->ext : (ulg)z->cext,
-                z->ext ? (z->cext ? "" : "local ") : "central ");
+          sprintf(errbuf,
+                  "local extra (%ld bytes) != central extra (%ld bytes): ",
+                  (ulg)z->ext, (ulg)z->cext);
+          if (noisy) fprintf(stderr, "\tzip info: %s%s\n", errbuf, z->zname);
         }
+#if (!defined(RISCOS) && !defined(CMS_MVS))
+        /* in noisy mode, extra field sizes are always reported */
+        else if (noisy)
+#else /* RISCOS || CMS_MVS */
+/* avoid warnings for zipfiles created on the same type of OS system! */
+/* or, was this warning really intended (eg. OS/2)? */
+        /* Only give info if extra bytes were added by another system */
+        else if (noisy && ((z->vem >> 8) != (OS_CODE >> 8)))
+#endif /* ?(RISCOS || CMS_MVS) */
+        {
+            fprintf(stderr, "zip info: %s has %ld bytes of %sextra data\n",
+                    z->zname, z->ext ? (ulg)z->ext : (ulg)z->cext,
+                    z->ext ? (z->cext ? "" : "local ") : "central ");
+        }
+    }
 }
 
 /*
- * readzipfile2 is called with zip -F or zip -FF
+ * scanzipf_fix is called with zip -F or zip -FF
  * read the file from front to back and pick up the pieces
  * NOTE: there are still checks missing to see if the header
  *       that was found is *VALID*
  */
-local int readzipfile2()
+local int scanzipf_fix(f)
+  FILE *f;                      /* zip file */
 /*
-   The name of the zip file is pointed to by the global "zipfile".  The
-   globals zfiles, zcount, zcomlen, zcomment, and zsort are filled in.
+   The name of the zip file is pointed to by the global "zipfile".  The globals
+   zipbeg, cenbeg, zfiles, zcount, zcomlen, zcomment, and zsort are filled in.
    Return an error code in the ZE_ class.
 */
 {
-  ulg a = 0L;           /* attributes returned by filetime() */
-  char b[CENHEAD];      /* buffer for central headers */
-  FILE *f;              /* zip file */
-  ush flg;              /* general purpose bit flag */
-  int m;                /* mismatch flag */
-  extent n;             /* length of name */
-  ulg p;                /* current file offset */
-  ulg s;                /* size of data, start of central */
-  struct zlist far * far *x;    /* pointer last entry's link */
-  struct zlist far *z;  /* current zip entry structure */
+    ulg a = 0L;                 /* attributes returned by filetime() */
+    char b[CENHEAD];            /* buffer for central headers */
+    ush flg;                    /* general purpose bit flag */
+    int m;                      /* mismatch flag */
+    extent n;                   /* length of name */
+    ulg p;                      /* current file offset */
+    ulg s;                      /* size of data, start of central */
+    struct zlist far * far *x;  /* pointer last entry's link */
+    struct zlist far *z;        /* current zip entry structure */
 
-  /* Initialize zip file info */
-  zipbeg = 0;
-  zfiles = NULL;                        /* Points to first header */
-  zcomlen = 0;                          /* zip file comment length */
-
-  /* If zip file exists, read headers and check structure */
-#ifdef VMS
-  if (zipfile == NULL || !(*zipfile) || !strcmp(zipfile, "-"))
-    return ZE_OK;
-  {
-    int rtype;
-
-    if ((VMSmunch(zipfile, GET_RTYPE, (char *)&rtype) == RMS$_NORMAL) &&
-        (rtype == FAT$C_VARIABLE)) {
-      fprintf(stderr,
-     "\n     Error:  zipfile is in variable-length record format.  Please\n\
-     run \"bilf b %s\" to convert the zipfile to fixed-length\n\
-     record format.\n\n", zipfile);
-      return ZE_FORM;
-    }
-  }
-  if ((f = fopen(zipfile, FOPR)) != NULL)
-#else /* !VMS */
-  if (zipfile != NULL && *zipfile && strcmp(zipfile, "-") &&
-      (f = fopen(zipfile, FOPR)) != NULL)
-#endif /* ?VMS */
-  {
     /* Get any file attribute valid for this OS, to set in the central
      * directory when fixing the archive:
      */
@@ -408,7 +430,6 @@ local int readzipfile2()
 #endif
     x = &zfiles;                        /* first link */
     p = 0;                              /* starting file offset */
-    zcount = 0;                         /* number of files */
 #ifdef HANDLE_AMIGA_SFX
     amiga_sfx_offset = 0L;
 #endif
@@ -448,7 +469,8 @@ local int readzipfile2()
     /* Read local headers */
     while (LG(b) == LOCSIG)
     {
-      if ((z = (struct zlist far *)farmalloc(sizeof(struct zlist))) == NULL)
+      if ((z = (struct zlist far *)farmalloc(sizeof(struct zlist))) == NULL ||
+          zcount + 1 < zcount)
         return ZE_MEM;
       if (fread(b, LOCHEAD, 1, f) != 1) {
           farfree((zvoid far *)z);
@@ -456,7 +478,7 @@ local int readzipfile2()
       }
 
       z->ver = SH(LOCVER + b);
-      z->vem = dosify ? 20 : OS_CODE + Z_MAJORVER * 10 + Z_MINORVER;
+      z->vem = (ush)(dosify ? 20 : OS_CODE + Z_MAJORVER * 10 + Z_MINORVER);
       z->dosflag = dosify;
       flg = z->flg = z->lflg = SH(LOCFLG + b);
       z->how = SH(LOCHOW + b);
@@ -487,7 +509,7 @@ local int readzipfile2()
       /* Read file name and extra field and skip data */
       if (n == 0)
       {
-        sprintf(errbuf, "%ld", (ulg)zcount + 1);
+        sprintf(errbuf, "%lu", (ulg)zcount + 1);
         zipwarn("zero-length name for entry #", errbuf);
 #ifndef DEBUG
         return ZE_FORM;
@@ -503,7 +525,7 @@ local int readzipfile2()
       /* If there is an extended local header, s is either 0 or
        * the correct compressed size.
        */
-      z->iname[n] = 0;                  /* terminate name */
+      z->iname[n] = '\0';               /* terminate name */
       z->zname = in2ex(z->iname);       /* convert to external name */
       if (z->zname == NULL)
         return ZE_MEM;
@@ -596,90 +618,42 @@ local int readzipfile2()
     }
 
     cenbeg = s;
-    zcomlen = 0;
 
     if (zipbeg && noisy)
       fprintf(mesg, "%s: adjusting offsets for a preamble of %lu bytes\n",
               zipfile, zipbeg);
 
-    /* Done with zip file for now */
-    fclose(f);
-
-    /* If one or more files, sort by name */
-    if (zcount)
-    {
-      if ((x = zsort =
-          (struct zlist far **)malloc(zcount * sizeof(struct zlist far *))) ==
-          NULL)
-        return ZE_MEM;
-      for (z = zfiles; z != NULL; z = z->nxt)
-        *x++ = z;
-      qsort((char *)zsort, zcount, sizeof(struct zlist far *), zqcmp);
-    }
-  }
-  return ZE_OK;
+    return ZE_OK;
 }
 
 #endif /* !UTIL */
 
 /*
- * readzipfile starts searching for the End Signature at the end of the file
+ * scanzipf_reg starts searching for the End Signature at the end of the file
  * The End Signature points to the Central Directory Signature which points
  * to the Local Directory Signature
  * XXX probably some more consistency checks are needed
  */
-int readzipfile()
+local int scanzipf_reg(f)
+  FILE *f;                      /* zip file */
 /*
-   The name of the zip file is pointed to by the global "zipfile".  The
-   globals zfiles, zcount, zcomlen, zcomment, and zsort are filled in.
+   The name of the zip file is pointed to by the global "zipfile".  The globals
+   zipbeg, cenbeg, zfiles, zcount, zcomlen, zcomment, and zsort are filled in.
    Return an error code in the ZE_ class.
 */
 {
-  char b[CENHEAD];      /* buffer for central headers */
-  FILE *f;              /* zip file */
-  ush flg;              /* general purpose bit flag */
-  int m;                /* mismatch flag */
-  extent n;             /* length of name */
-  struct zlist far * far *x;    /* pointer last entry's link */
-  struct zlist far *z;  /* current zip entry structure */
-  char *t;              /* temporary pointer */
-  char far *u;          /* temporary variable */
-  int found;
-  char *buf;            /* temp buffer for reading zipfile */
-  long deltaoff;
+    char b[CENHEAD];            /* buffer for central headers */
+    ush flg;                    /* general purpose bit flag */
+    int m;                      /* mismatch flag */
+    extent n;                   /* length of name */
+    struct zlist far * far *x;  /* pointer last entry's link */
+    struct zlist far *z;        /* current zip entry structure */
+    char *t;                    /* temporary pointer */
+    char far *u;                /* temporary variable */
+    int found;
+    char *buf;                  /* temp buffer for reading zipfile */
+    long deltaoff;
 
-#ifndef UTIL
-  if (fix && !adjust)
-     return readzipfile2();
-#endif
-
-  /* Initialize zip file info */
-  zipbeg = 0;
-  zfiles = NULL;                        /* Points to first header */
-  zcomlen = 0;                          /* zip file comment length */
-
-  /* If zip file exists, read headers and check structure */
-#ifdef VMS
-  if (zipfile == NULL || !(*zipfile) || !strcmp(zipfile, "-"))
-    return ZE_OK;
-  {
-    int rtype;
-
-    if ((VMSmunch(zipfile, GET_RTYPE, (char *)&rtype) == RMS$_NORMAL) &&
-        (rtype == FAT$C_VARIABLE)) {
-      fprintf(stderr,
-     "\n     Error:  zipfile is in variable-length record format.  Please\n\
-     run \"bilf b %s\" to convert the zipfile to fixed-length\n\
-     record format.\n\n", zipfile);
-      return ZE_FORM;
-    }
-  }
-  if ((f = fopen(zipfile, FOPR)) != NULL)
-#else /* !VMS */
-  if (zipfile != NULL && *zipfile && strcmp(zipfile, "-") &&
-      (f = fopen(zipfile, FOPR)) != NULL)
-#endif /* ?VMS */
-  {
     buf = malloc(4096 + 4);
     if (buf == NULL)
       return ZE_MEM;
@@ -733,7 +707,8 @@ int readzipfile()
     if (!found && zipbeg > 0) {
       size_t s;
 
-      rewind(f);
+      fseek(f, 0L, SEEK_SET);
+      clearerr(f);
       s = fread(buf, 1, (size_t) zipbeg, f);
       buf[s] = t[1];
       buf[s + 1] = t[2];
@@ -795,7 +770,6 @@ int readzipfile()
     }
 
     x = &zfiles;                        /* first link */
-    zcount = 0;                         /* number of files */
 
     if (fread(b, 4, 1, f) != 1)
       return ferror(f) ? ZE_READ : ZE_EOF;
@@ -847,7 +821,7 @@ int readzipfile()
           (z->cext && fread(z->cextra, z->cext, 1, f) != 1) ||
           (z->com && fread(z->comment, z->com, 1, f) != 1))
         return ferror(f) ? ZE_READ : ZE_EOF;
-      z->iname[z->nam] = 0;                  /* terminate name */
+      z->iname[z->nam] = '\0';                  /* terminate name */
 #ifdef EBCDIC
       if (z->com)
          memtoebc(z->comment, z->comment, z->com);
@@ -1032,22 +1006,134 @@ int readzipfile()
       }
     }
 #endif /* HANDLE_AMIGA_SFX */
+    return ZE_OK;
+}
+
+
+/*
+ * readzipfile initializes the global variables that hold the zipfile
+ * directory info and opens the zipfile. For the actual zipfile scan,
+ * the subroutine scanzipf_reg() or scanzipf_fix() is called,
+ * depending on the mode of operation (regular processing, or zipfix mode).
+ */
+int readzipfile()
+/*
+   The name of the zip file is pointed to by the global "zipfile".
+   The globals zipbeg, zfiles, zcount, and zcomlen are initialized.
+   Return an error code in the ZE_ class.
+*/
+{
+  FILE *f;              /* zip file */
+  int retval;           /* return code */
+  int readable;         /* 1 if zipfile exists and is readable */
+
+  /* Initialize zip file info */
+  zipbeg = 0;
+  zfiles = NULL;                        /* Points to first header */
+  zcount = 0;                           /* number of files */
+  zcomlen = 0;                          /* zip file comment length */
+  retval = ZE_OK;
+  f = NULL;                             /* shut up some compilers */
+
+  /* If zip file exists, read headers and check structure */
+#ifdef VMS
+  if (zipfile == NULL || !(*zipfile) || !strcmp(zipfile, "-"))
+    return ZE_OK;
+  {
+    int rtype;
+
+    if ((VMSmunch(zipfile, GET_RTYPE, (char *)&rtype) == RMS$_NORMAL) &&
+        (rtype == FAT$C_VARIABLE)) {
+      fprintf(stderr,
+     "\n     Error:  zipfile is in variable-length record format.  Please\n\
+     run \"bilf b %s\" to convert the zipfile to fixed-length\n\
+     record format.\n\n", zipfile);
+      return ZE_FORM;
+    }
+  }
+  readable = ((f = fopen(zipfile, FOPR)) != NULL);
+#else /* !VMS */
+  readable = (zipfile != NULL && *zipfile && strcmp(zipfile, "-") &&
+              (f = fopen(zipfile, FOPR)) != NULL);
+#endif /* ?VMS */
+#ifdef MVS
+  /* Very nasty special case for MVS.  Just because the zipfile has been
+   * opened for reading does not mean that we can actually read the data.
+   * Typical JCL to create a zipfile is
+   *
+   * //ZIPFILE  DD  DISP=(NEW,CATLG),DSN=prefix.ZIP,
+   * //             SPACE=(CYL,(10,10))
+   *
+   * That creates a VTOC entry with an end of file marker (DS1LSTAR) of zero.
+   * Alas the VTOC end of file marker is only used when the file is opened in
+   * append mode.  When a file is opened in read mode, the "other" end of file
+   * marker is used, a zero length data block signals end of file when reading.
+   * With a brand new file which has not been written to yet, it is undefined
+   * what you read off the disk.  In fact you read whatever data was in the same
+   * disk tracks before the zipfile was allocated.  You would be amazed at the
+   * number of application programmers who still do not understand this.  Makes
+   * for interesting and semi-random errors, GIGO.
+   *
+   * Newer versions of SMS will automatically write a zero length block when a
+   * file is allocated.  However not all sites run SMS or they run older levels
+   * so we cannot rely on that.  The only safe thing to do is close the file,
+   * open in append mode (we already know that the file exists), close it again,
+   * reopen in read mode and try to read a data block.  Opening and closing in
+   * append mode will write a zero length block where DS1LSTAR points, making
+   * sure that the VTOC and internal end of file markers are in sync.  Then it
+   * is safe to read data.  If we cannot read one byte of data after all that,
+   * it is a brand new zipfile and must not be read.
+   */
+  if (readable)
+  {
+    char c;
+    fclose(f);
+    /* append mode */
+    if ((f = fopen(zipfile, "ab")) == NULL) {
+      ZIPERR(ZE_OPEN, zipfile);
+    }
+    fclose(f);
+    /* read mode again */
+    if ((f = fopen(zipfile, FOPR)) == NULL) {
+      ZIPERR(ZE_OPEN, zipfile);
+    }
+    if (fread(&c, 1, 1, f) != 1) {
+      /* no actual data */
+      readable = 0;
+      fclose(f);
+    }
+    else{
+      fseek(f, 0, SEEK_SET);  /* at least one byte in zipfile, back to the start */
+    }
+  }
+#endif /* MVS */
+  if (readable)
+  {
+#ifndef UTIL
+    retval = (fix && !adjust) ? scanzipf_fix(f) : scanzipf_reg(f);
+#else
+    retval = scanzipf_reg(f);
+#endif
+
     /* Done with zip file for now */
     fclose(f);
 
     /* If one or more files, sort by name */
     if (zcount)
     {
-      if ((x = zsort =
-          (struct zlist far **)malloc(zcount * sizeof(struct zlist far *))) ==
-          NULL)
+      struct zlist far * far *x;    /* pointer into zsort array */
+      struct zlist far *z;          /* pointer into zfiles linked list */
+      extent zl_size = zcount * sizeof(struct zlist far *);
+
+      if (zl_size / sizeof(struct zlist far *) != zcount ||
+          (x = zsort = (struct zlist far **)malloc(zl_size)) == NULL)
         return ZE_MEM;
       for (z = zfiles; z != NULL; z = z->nxt)
         *x++ = z;
       qsort((char *)zsort, zcount, sizeof(struct zlist far *), zqcmp);
     }
   }
-  return ZE_OK;
+  return retval;
 }
 
 
@@ -1111,7 +1197,7 @@ FILE *f;                /* file to write to */
 #ifdef EBCDIC
   if (z->com)
     memtoasc(z->comment, z->comment, z->com);
-#endif
+#endif /* EBCDIC */
   if (fwrite(z->iname, 1, z->nam, f) != z->nam ||
       (z->cext && fwrite(z->cextra, 1, z->cext, f) != z->cext) ||
       (z->com && fwrite(z->comment, 1, z->com, f) != z->com))
@@ -1287,7 +1373,7 @@ iztimes *z_utim;                /* return storage: atime, mtime, ctime */
            if (flags & EB_UT_FL_CTIME) {
               if ((eb_idx+4) <= eb_len) {
                  z_utim->ctime = LG((EB_HEADSIZE+eb_idx) + ef_buf);
-                 eb_idx += 4;
+                 /* eb_idx += 4; */  /* superfluous for now ... */
                  Trace((stderr,"  Unix EF cretime = %ld\n", z_utim->ctime));
               } else {
                  flags &= ~EB_UT_FL_CTIME;
@@ -1317,6 +1403,10 @@ iztimes *z_utim;                /* return storage: atime, mtime, ctime */
         }
         break;
 
+      case EF_THEOS:
+/*      printf("Not implemented yet\n"); */
+        break;
+
       default:
         break;
     }
@@ -1334,6 +1424,10 @@ iztimes *z_utim;
 {
   int r;
 
+#ifdef IZ_CHECK_TZ
+  if (!zp_tz_is_valid) return 0;
+#endif
+
   /* First, scan local extra field. */
   r = ef_scan_ut_time(z->extra, z->ext, FALSE, z_utim);
 
@@ -1348,8 +1442,9 @@ iztimes *z_utim;
 #endif /* USE_EF_UT_TIME */
 
 
-local void cutpath(p)
+local void cutpath(p, delim)
 char *p;                /* path string */
+int delim;              /* path component separator char */
 /* Cut the last path component off the name *p in place.
  * This should work on both internal and external names.
  */
@@ -1357,10 +1452,10 @@ char *p;                /* path string */
   char *r;              /* pointer to last path delimiter */
 
 #ifdef VMS                      /* change [w.x.y]z to [w.x]y.DIR */
-  if ((r = strrchr(p, ']')) != NULL)
+  if ((r = MBSRCHR(p, ']')) != NULL)
   {
     *r = 0;
-    if ((r = strrchr(p, '.')) != NULL)
+    if ((r = MBSRCHR(p, '.')) != NULL)
     {
       *r = ']';
       strcat(r, ".DIR;1");     /* this assumes a little padding--see PAD */
@@ -1368,13 +1463,13 @@ char *p;                /* path string */
       *p = 0;
     }
   } else {
-    if ((r = strrchr(p, '/')) != NULL)
+    if ((r = MBSRCHR(p, delim)) != NULL)
       *r = 0;
     else
       *p = 0;
   }
 #else /* !VMS */
-  if ((r = strrchr(p, '/')) != NULL)
+  if ((r = MBSRCHR(p, delim)) != NULL)
     *r = 0;
   else
     *p = 0;
@@ -1397,7 +1492,7 @@ int trash()
     if (z->mark == 1 || z->trash)
     {
       z->mark = 1;
-      if (z->iname[z->nam - 1] != '/') { /* don't unlink directory */
+      if (z->iname[z->nam - 1] != (char)0x2f) { /* don't unlink directory */
         if (verbose)
           fprintf(mesg, "zip diagnostic: deleting file %s\n", z->name);
         if (destroy(z->name)) {
@@ -1407,12 +1502,13 @@ int trash()
          * necessary only with the -D option.
          */
         if (!dirnames) {
-          cutpath(z->name);  /* XXX wrong ??? */
-          cutpath(z->iname);
-          if (z->iname[0] != '\0') {
-            strcat(z->iname, "/");
-          }
+          cutpath(z->name, '/');  /* XXX wrong ??? */
+          cutpath(z->iname, 0x2f); /* 0x2f = ascii['/'] */
           z->nam = strlen(z->iname);
+          if (z->nam > 0) {
+            z->iname[z->nam - 1] = (char)0x2f;
+            z->iname[z->nam++] = '\0';
+          }
           if (z->nam > 0) n++;
         }
       } else {
@@ -1430,7 +1526,7 @@ int trash()
       return ZE_MEM;
     n = 0;
     for (z = zfiles; z != NULL; z = z->nxt) {
-      if (z->mark && z->nam > 0 && z->iname[z->nam - 1] == '/'
+      if (z->mark && z->nam > 0 && z->iname[z->nam - 1] == (char)0x2f /* '/' */
           && (n == 0 || strcmp(z->name, s[n-1]->name) != 0)) {
         s[n++] = z;
       }
